#
# Script to implement the Spruce Budworm IPS in Tcl/SDB.
# Tom Moore, September 6, 1995
#
# Execute the tcl script 'convert.tcl' prior to running this file.
#
# There are three parts in this file:
#
# 1) Define lists of volume components and specify how these
#    will be aggregated.
#
# 2) Load the data files, and define expressions interpolate
#    the yield curves and calculate stand volumes.
#
# 3) Define procedures that do the projections.  Forecasts
#    can be conducted by exectuing these procedures.
#

############################################################
#
# Define some characteristics about this inventory.
# All of the characteristics listed below have been obtained
# from the various ips.* files.
#
############################################################

#
# Known volume components not impacted by budworm (obtained from 
# definitions in the curveset file).  
#
set volumeComps {PINLOG PINSTD PINPLP IH_PLP OS_PLP TH_LOG TH_PLP WP_LOG EC_LOG PINOSZ WP_PLP EC_PLP M3_TR}

#
# Component volumes impacted by outbreaks (obtained from the
# volume impact link)
#
set BFImpact {FIRPLP FIRSTD FIRLOG FIROSZ}
set SPImpact {SPRPLP SPRSTD SPRLOG SPROSZ}

#
# Primary and secondary products and volumes (obtained from the
# curve composition file)
#
set PrimaryVolume {SPRPLP SPRLOG SPROSZ FIRPLP FIRLOG FIROSZ PINPLP PINLOG PINOSZ}
set SecondaryVolume {IH_PLP TH_PLP}
set PrimaryProduct {SPRSTD FIRSTD PINSTD}
set SecondaryProduct {TH_LOG}

#
# Hardwood volume components (hardwood volume component file)
#
set Hardwood {IH_PLP TH_PLP TH_LOG}

#
# Funa group volumes (from the funa components file)
#
set OSVolume {EC_PLP EC_LOG OS_PLP}
set IHVolume {IH_PLP}
set SPVolume {SPRPLP SPRLOG SPROSZ}
set BFVolume {FIRPLP FIRLOG FIROSZ}
set THVolume {TH_PLP}
set PIVolume {WP_PLP WP_LOG PINPLP PINLOG PINOSZ}

set AllVolume {OSVolume IHVolume SPVolume BFVolume THVolume PIVolume}

#
# Names of all of the lists defined so far.  This will be used later
# to automatically generate expressions based on the lists of
# volume groups.
#
set Aggregate {PrimaryVolume SecondaryVolume PrimaryProduct SecondaryProduct BFImpact SPImpact Hardwood OSVolume IHVolume SPVolume BFVolume THVolume PIVolume AllVolume}


############################################################
#
# Initialize the environment of the simulation by reading in 
# the basic datasets.
#
############################################################

puts "Begin loading the ips datasets"

#
# Read in the primary datasets 
#

#
# First open the starting inventory file.  This file contains
# one record for each stand in the forest inventory.  This file
# will be used as the base relation, all other files will be joined
# to it in one-to-one relations.  When the database is opened it
# will be given a symbolic name of 'ips'.  This will create an
# 'ips' command that can be used to manipulate the database, and
# an 'ips' array variable that can be used to set and retrieve 
# database varaibles.
#
puts "open inventory"
sdb open ips stand.inv

#
# Join in the stand class link file.  This will add the column
# 'InitialClass'.  Create a column in the base relation call 'Class',
# and initialize it with 'InitialClass'.
#
puts "Join stand-class link file"
ips join load scl stand.lu
ips allocate Class CODE
ips calculate {Class=InitialClass}

#
# Join in the class file.  This relation will be based on the column
# 'Class' that was defined in the previous step.  Create columns in the
# base relation for 'Age' and 'Curveset', and initialize them with
# values from the class relation.
#
puts "join class file"
ips join load class class.lu
ips allocate Age NUMBER
ips calculate {Age=AGE}
ips allocate Curveset CODE
ips calculate {Curveset=InitialCurveset}

#
# Load curveset.  The curvesets have been normalized into two parts:"
# the curveset descriptors include the entry age, breakup age, a type"
# name, and breakup curveset; the curveset 'curve' files include only 
# the yield curve data.  These files will be joined by the field
# 'Curveset' that was defined in the previous step.
#
puts "join curve set descriptor"
ips join load set set.lu

puts "load yield curves"
ips curve load v curve.yld

#
# Load in the harvest schedule file.  The join is based on the
# value of the 'STAND' variable in the base relation.  This relation
# adds the variable 'PERIOD' which describes the future 5-year period
# when the stand is scheduled for harvest.
#
puts "join harvest schedule"
ips join load schedule schedule.dat


#
# Load in the impact matrix.  The impact matrix has been converted
# into a yield file in order to allow interpolation between
# vertices.  Before loading the impact curves there must exist
# definitions for the fields that will be used to index the
# curves.  In this case the fields are Maturity, Severity, and
# HdwdCode.  A field named YrSinceOB will be used to index the 
# interpolation along the curve.
#

#
# DNRE defined stand development stages
#
ips define Maturity {Age?(<,5,"C"):(<=,25,"R"):(<=,45,"Y"):(<=,70,"I"):(<=,100,"M"):"O"}
ips define HdwdCode {"none"}
ips define Severity {"none"}
ips define YrSinceOB {0}

ips curve load impact matrix.yld


#
# Define the volume equations, one for each defined in the curveset
# Most species are simply interpolated along a yield curve.  The
# BP and SP volume components are suceptable to budworm, and
# are multiplied by a impact factor (<=1.0) that varies according
# to time since impact, maturity, severity and hardwood composition
# of the stand.  Impact factors are defined in the impact matrix
# curve set read in previously.
#
foreach spp $volumeComps {
    ips define $spp "interpolate(\"v\",\"Age\",\"$spp\",Age)"
}
foreach spp $BFImpact {
    ips define $spp "interpolate(\"v\",\"Age\",\"$spp\",Age)*interpolate(\"impact\",\"YrSinceOB\",\"BF_Impact\",YrSinceOB)"
}
foreach spp $SPImpact {
    ips define $spp "interpolate(\"v\",\"Age\",\"$spp\",Age)*interpolate(\"impact\",\"YrSinceOB\",\"BF_Impact\",YrSinceOB)"
}

#
# Define expressions which aggregate volume components according to
# the lists entered at the beginning of this file.  For each
# list construct 2 variable definitions:  the first is simply
# the sum of all components in the list; the second is the sum
# multiplied by the area of the stand record.
#
foreach product $Aggregate {
    set expression [lindex [set $product] 0]
    foreach component [lrange [set $product] 1 end] {
	lappend expression "+" $component
    }
    puts "Defining $expression"
    ips define $product $expression
    ips define Total[set product] "$product * Area"
}


#
# Define fields which determine the softwood volume and also the
# softwood and hardwood percentage volumes.  
#
ips define HdwdPct  {(Hardwood>0)?(==,1,Hardwood/AllVolume):0}
ips define Softwood {AllVolume-Hardwood}
ips define SfwdPct  {(Softwood>0)?(==,1,Softwood/AllVolume):0}

#
# Perform a calculation to determine a classification of the 
# amount of hardwood volume in a stand.  The result is used as
# a factor in the impact matrix (presence of hardwood mitigates
# impact).
#
ips calculate {HdwdCode=HdwdPct?(>,0.2,"some"):"none"}

#
# Define an expression which computes the funa code.  This
# is a translation of the 'C' code written by Ian DeMerchant.
# It appears as though there are some inconsistencies in this
# function.
#
puts "define funa codes"
ips define Funa {AllVolume?(==,0,"CC"): \
	SfwdPct? \
          (>=,0.75,\
             AllVolume*0.5? \
	       (<,SPVolume,"SPBF"): \
               (<,BFVolume,"BFSP"): \
               (<,PIVolume,"PINE"): \
               (<,OSVolume,"OTSW"): \
               OSVolume?(>,Softwood*0.5,"OTSW"): \
               "Uknown SW"): \
	  (>=,0.5,THVolume? \
            (>=,IHVolume,AllVolume*0.5?(>,BFVolume,"SPTH"):"BFTH"): \
             AllVolume*0.5?(>,BFVolume,"SPIH"):"BFIH"): \
	  (>=,0.25,THVolume? \
	     (>=,IHVolume,(Softwood*.5)?(>,BFVolume,"THSP"):"THBF"): \
             Softwood*.5?(>,BFVolume,"IHSP"):"IHBF"): \
	  THVolume? \
            (>,IHVolume,THVolume?(>=,Hardwood*.75,"TOHW"): \
                                 (>,Hardwood*.5,"THIH"): \
                                  "Unknown TH"): \
	    IHVolume?(>=,Hardwood*.75,"INHW"): \
                     (>,Hardwood*.5,"IHTH"): \
                     "Unknown IH"}



############################################################
#
# Define procedures to implement the projections.
#
############################################################

#
# Define a procedure to step through time
#
proc ips_forecast {baseYear OBYear endYear Severity} {

    global ips
    set ips(Severity) $Severity

    puts "Simulate the advance of the forest inventory from $baseYear to $endYear"
    puts "The outbreak will occur in $OBYear and will be of a $ips(Severity) level of severity"

    ips_reset

    #
    # Loop ahead through time
    #
    puts "Starting simulation loop"
    set period 0
    for {set date $baseYear} {[incr date 5] <= $endYear} {} {
	
	puts "update age"
	incr period
	ips calculate {Age=Age+5}

	#
	# Determine the time since outbreak (if it has occured yet)
	#
	if {[set YrSinceOB [expr $date - $OBYear]] < 0} \
		then {set YrSinceOB 0} \
		else {puts "It is now $YrSinceOB years since the outbreak started"}
	# update the HSG variable
	set ips(YrSinceOB) $YrSinceOB
	
	#
	# Update stands harvested this period
	#
	puts "update harvest"
	ips select set
	ips select and "PERIOD==\"$period\""
	ips calculate "Age=0"
	
	#
	# Update stands that exceed the breakup age and start them
	# along new curves
	#
	puts "regenerate stands"
	ips select set
	ips select and {Age > BreakupAge}
	ips calculate {Age=EntryAge}
	ips calculate {Curveset=BreakupCurve}
	ips select set
	
	#
	# The inventory is now up to date.  Compute any summary information
	# here.
	#
	puts "calculate summary for $date"
	puts "Summary of volumes by funa and maturity class"
	puts [ips summary Funa Maturity TotalAllVolume]
	if {$YrSinceOB > 0 && \
		($Severity == "Severe" || $Severity == "Moderate")} \
		then {
	    puts " "
	    puts "If the outbreak had not occured, the volumes would have been"
	    set ips(Severity) None
	    puts [ips summary Funa Maturity TotalAllVolume]
	    set ips(Severity) $Severity
	}	    
    }
}

#
# Procedure to restore the database to initial conditions
#
proc ips_reset {} {
    puts "Restore database to initial conditions"
    ips select set
    ips calculate {Class=InitialClass}
    ips calculate {Curveset=InitialCurveset}
    ips calculate {Age=AGE}
}


############################################################
#
# End of IPS definitions
#
############################################################


puts "The database has been loaded." 
puts " "
puts "To begin forecasting enter the command:"
puts "  ips_forecast 'StartYear' 'OBYear' 'EndYear' 'Severity'"
puts "where"
puts "  StartYear is the starting year for the forecast,"
puts "  OBYear is the year when the outbreak is scheduled to occur,"
puts "  EndYear is the ending year for the forecast, and"
puts "  Severity is the severity of the outbreak, which can be one"
puts "  of 'Moderate', 'Severe', or any other value which indicates"
puts "  no outbreak."
puts " "
puts "To reset the simulation database to initial conditions use the"
puts "command:"
puts "  ips_reset"
puts " "

